import 'package:flutter/material.dart';

import 'circular_loading.dart';

enum ButtonType {
  raised,
  flat,
  outline,
  text,
}

class CustomButtonStyle {
  final Color color;
  final Color textColor;
  final ButtonType buttonType;
  const CustomButtonStyle(
      {@required this.buttonType, this.color, this.textColor});
}

class Button extends StatelessWidget {
  final CustomButtonStyle buttonStyle;
  final Function onPressed;
  final String text;
  final String loadingText;
  final double fontSizeFactor;
  final double fontSizeDelta;
  final IconData icon;
  final TextAlign textAlign;
  final MainAxisAlignment align;
  const Button({
    @required this.onPressed,
    this.buttonStyle,
    this.text,
    this.loadingText,
    this.fontSizeFactor = 1.0,
    this.fontSizeDelta = 0.0,
    this.icon,
    this.textAlign = TextAlign.center,
    this.align = MainAxisAlignment.center,
    Key key,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    final style =
        buttonStyle ?? CustomButtonStyle(buttonType: ButtonType.raised);
    switch (style.buttonType) {
      case ButtonType.text:
        return _buildButton(style, _buildTextButton);
      case ButtonType.outline:
        return _buildButton(style, _buildOutlineButton);
      case ButtonType.flat:
        return _buildButton(style, _buildFlatButton);
      case ButtonType.raised:
      default:
        return _buildButton(style, _buildRaisedButton);
    }
  }

  Widget _buildButton(
      CustomButtonStyle style,
      Function(CustomButtonStyle style, Function onPressed, bool loading,
              BuildContext context)
          buttonBuilder) {
    var loading = false;
    return StatefulBuilder(builder: (_context, setState) {
      return buttonBuilder(
          style,
          onPressed == null
              ? null
              : () async {
                  if (loading) return;
                  setState(() {
                    loading = true;
                  });
                  await onPressed();
                  setState(() {
                    loading = false;
                  });
                },
          loading,
          _context);
    });
  }

  Widget _buildTextButton(CustomButtonStyle style, Function onPressed,
      bool loading, BuildContext context) {
    final theme = Theme.of(context);
    final textWidget = Text(
      (loading && loadingText != null && loadingText.isNotEmpty)
          ? loadingText
          : text,
      textAlign: textAlign,
      style: theme.textTheme.button.apply(
          color: onPressed != null ? style.textColor : Colors.black26,
          fontSizeFactor: fontSizeFactor,
          fontSizeDelta: fontSizeDelta),
    );
    return Row(
      mainAxisAlignment: align,
      children: [
        if (icon != null)
          Icon(icon,
              color: onPressed == null ? theme.disabledColor : style.textColor),
        TextButton(onPressed: onPressed, child: textWidget),
      ],
    );
  }

  Widget _buildOutlineButton(CustomButtonStyle style, Function onPressed,
      bool loading, BuildContext context) {
    final theme = Theme.of(context);
    final textWidget = Text(
      (loading && loadingText != null && loadingText.isNotEmpty)
          ? loadingText
          : text,
      textAlign: TextAlign.center,
      style: theme.textTheme.button.apply(
          color: onPressed != null ? style.textColor : Colors.black26,
          fontSizeFactor: fontSizeFactor,
          fontSizeDelta: fontSizeDelta),
    );

    return OutlineButton(
      onPressed: onPressed,
      color: style.color,
      child: Row(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          if (loading)
            CircularLoading(
              color: style.textColor,
            ),
          if (icon != null)
            Icon(icon,
                color:
                    onPressed == null ? theme.disabledColor : style.textColor),
          textWidget
        ],
      ),
    );
  }

  Widget _buildRaisedButton(CustomButtonStyle style, Function onPressed,
      bool loading, BuildContext context) {
    final theme = Theme.of(context);
    final textWidget = Text(
      (loading && loadingText != null && loadingText.isNotEmpty)
          ? loadingText
          : text,
      textAlign: TextAlign.center,
      style: theme.textTheme.button.apply(
          color: onPressed != null ? style.textColor : Colors.black26,
          fontSizeFactor: fontSizeFactor,
          fontSizeDelta: fontSizeDelta),
    );

    return RaisedButton(
      onPressed: onPressed,
      color: style.color,
      child: Row(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          if (loading)
            CircularLoading(
              color: style.textColor,
            ),
          if (icon != null)
            Icon(icon,
                color:
                    onPressed == null ? theme.disabledColor : style.textColor),
          textWidget
        ],
      ),
    );
  }

  Widget _buildFlatButton(CustomButtonStyle style, Function onPressed,
      bool loading, BuildContext context) {
    final theme = Theme.of(context);
    final textWidget = Text(
      (loading && loadingText != null && loadingText.isNotEmpty)
          ? loadingText
          : text,
      textAlign: TextAlign.center,
      style: theme.textTheme.button.apply(
          color: onPressed != null ? style.textColor : Colors.black26,
          fontSizeFactor: fontSizeFactor,
          fontSizeDelta: fontSizeDelta),
    );

    return FlatButton(
      onPressed: onPressed,
      color: style.color,
      child: Row(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          if (loading)
            CircularLoading(
              color: style.textColor,
            ),
          if (icon != null)
            Icon(icon,
                color:
                    onPressed == null ? theme.disabledColor : style.textColor),
          textWidget
        ],
      ),
    );
  }
}
