import 'dart:ui';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:mobile_ui/mobile_ui.dart';

const double _kTabHeight = 46.0;

class StepperBar extends StatefulWidget implements PreferredSizeWidget {
  const StepperBar(
      {Key key,
      this.controller,
      this.currentStepChange,
      this.length,
      this.indicatorHeight = 3.0,
      this.canTap,
      this.paddingFlactor = 6})
      : super(key: key);

  final int length;
  final TabController controller;
  final ValueChanged<int> currentStepChange;
  final Function canTap;
  final double indicatorHeight;
  final double paddingFlactor;

  @override
  Size get preferredSize {
    return Size.fromHeight(indicatorHeight + 20);
  }

  @override
  _StepperBarState createState() => _StepperBarState();
}

class _StepperBarState extends State<StepperBar> {
  TabController _controller;
  double _currentAnimationValue = 1;

  @override
  void initState() {
    super.initState();
  }

  // If the TabBar is rebuilt with a new tab controller, the caller should
  // dispose the old one. In that case the old controller's animation will be
  // null and should not be accessed.
  bool get _controllerIsValid => _controller?.animation != null;

  void _updateTabController() {
    final newController = widget.controller ?? DefaultTabController.of(context);
    assert(() {
      if (newController == null) {
        throw FlutterError('No TabController for ${widget.runtimeType}.\n'
            'When creating a ${widget.runtimeType}, you must either provide an explicit '
            'TabController using the "controller" property, or you must ensure that there '
            'is a DefaultTabController above the ${widget.runtimeType}.\n'
            'In this case, there was neither an explicit controller nor a default controller.');
      }
      return true;
    }());

    if (newController == _controller) return;

    if (_controllerIsValid) {
      _controller.animation.removeListener(_handleTabControllerAnimationTick);
      _controller.removeListener(_handleTabControllerTick);
    }
    _controller = newController;
    if (_controller != null) {
      _controller.animation.addListener(_handleTabControllerAnimationTick);
      _controller.addListener(_handleTabControllerTick);
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    assert(debugCheckHasMaterial(context));
    _updateTabController();
  }

  @override
  void didUpdateWidget(StepperBar oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.controller != oldWidget.controller) {
      _updateTabController();
    }
  }

  @override
  void dispose() {
    if (_controllerIsValid) {
      _controller.animation.removeListener(_handleTabControllerAnimationTick);
      _controller.removeListener(_handleTabControllerTick);
    }
    _controller = null;
    // We don't own the _controller Animation, so it's not disposed here.
    super.dispose();
  }

  void _handleTabControllerTick() {
    if (widget.currentStepChange != null) {
      widget.currentStepChange(_controller.index);
    }
  }

  void _handleTap(int index) {
    if (widget.canTap()) {
      assert(index >= 0 && index < widget.length);
      _controller.animateTo(index);
    }
  }

  @override
  Widget build(BuildContext context) {
    assert(debugCheckHasMaterialLocalizations(context));
    assert(() {
      if (_controller.length != widget.length) {
        throw FlutterError(
            'Controller\'s length property (${_controller.length}) does not match the \n'
            'number of tabs (${widget.length}) present in TabBar\'s tabs property.');
      }
      return true;
    }());
    if (_controller.length == 0) {
      return Container(
        height: _kTabHeight + widget.indicatorHeight,
      );
    }

    final wrappedTabs = <Widget>[];
    final tabCount = widget.length;
    final changingIndex = _currentAnimationValue.floor();
    for (var index = 0; index < tabCount; index += 1) {
      var tab;
      if (index < changingIndex) {
        tab = StepperProgressIndicator(
          indicatorHeight: widget.indicatorHeight,
          value: 1,
        );
      } else if (index > changingIndex) {
        tab = StepperProgressIndicator(
          indicatorHeight: widget.indicatorHeight,
          value: 0,
        );
      } else {
        tab = StepperProgressIndicator(
          indicatorHeight: widget.indicatorHeight,
          value: _currentAnimationValue - changingIndex,
        );
      }
      if (index > 0) {
        wrappedTabs.add(SizedBox(
          width: 4,
        ));
      }
      wrappedTabs.add(Expanded(
          child: InkWell(
              onTap: () {
                _handleTap(index);
              },
              child: Padding(
                child: tab,
                padding: EdgeInsets.symmetric(vertical: 10),
              ))));
    }
    return Padding(
        padding:
            EdgeInsets.symmetric(horizontal: uiSpacing(widget.paddingFlactor)),
        child: Row(
          crossAxisAlignment: CrossAxisAlignment.end,
          children: wrappedTabs,
        ));
  }

  void _handleTabControllerAnimationTick() {
    setState(() {
      _currentAnimationValue = _controller.animation.value + 1;
    });
  }
}

class StepperProgressIndicator extends StatefulWidget {
  const StepperProgressIndicator(
      {Key key,
      this.value,
      this.backgroundColor,
      this.valueColor,
      this.semanticsLabel,
      this.semanticsValue,
      this.indicatorHeight})
      : super(key: key);
  final double indicatorHeight;
  final double value;
  final Color backgroundColor;
  final Animation<Color> valueColor;
  final String semanticsLabel;
  final String semanticsValue;

  Color _getBackgroundColor(BuildContext context) =>
      backgroundColor ?? Colors.grey[200];

  Color _getValueColor(BuildContext context) =>
      valueColor?.value ?? Theme.of(context).accentColor;

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties.add(PercentProperty('value', value,
        showName: false, ifNull: '<indeterminate>'));
  }

  Widget _buildSemanticsWrapper({
    @required BuildContext context,
    @required Widget child,
  }) {
    var expandedSemanticsValue = semanticsValue;
    if (value != null) {
      expandedSemanticsValue ??= '${(value * 100).round()}%';
    }
    return Semantics(
      label: semanticsLabel,
      value: expandedSemanticsValue,
      child: child,
    );
  }

  @override
  _StepperProgressIndicatorState createState() =>
      _StepperProgressIndicatorState();
}

const int _kIndeterminateLinearDuration = 1800;

class _StepperProgressIndicatorState extends State<StepperProgressIndicator>
    with SingleTickerProviderStateMixin {
  AnimationController _controller;

  @override
  void initState() {
    super.initState();
    _controller = AnimationController(
      duration: const Duration(milliseconds: _kIndeterminateLinearDuration),
      vsync: this,
    );
    if (widget.value == null) _controller.repeat();
  }

  @override
  void didUpdateWidget(StepperProgressIndicator oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.value == null && !_controller.isAnimating) {
      _controller.repeat();
    } else if (widget.value != null && _controller.isAnimating) {
      _controller.stop();
    }
  }

  @override
  void dispose() {
    _controller.dispose();
    super.dispose();
  }

  Widget _buildIndicator(BuildContext context, double animationValue,
      TextDirection textDirection) {
    return widget._buildSemanticsWrapper(
      context: context,
      child: Container(
        height: widget.indicatorHeight,
        constraints: const BoxConstraints(
          minWidth: double.infinity,
          minHeight: 2,
        ),
        child: CustomPaint(
          painter: _StepperProgressIndicatorPainter(
            backgroundColor: widget._getBackgroundColor(context),
            valueColor: widget._getValueColor(context),
            value: widget.value,
            // may be null
            animationValue: animationValue,
            // ignored if widget.value is not null
            textDirection: textDirection,
          ),
        ),
      ),
    );
  }

  @override
  Widget build(BuildContext context) {
    final textDirection = Directionality.of(context);

    if (widget.value != null) {
      return _buildIndicator(context, _controller.value, textDirection);
    }

    return AnimatedBuilder(
      animation: _controller.view,
      builder: (context, child) {
        return _buildIndicator(context, _controller.value, textDirection);
      },
    );
  }
}

class _StepperProgressIndicatorPainter extends CustomPainter {
  const _StepperProgressIndicatorPainter({
    this.backgroundColor,
    this.valueColor,
    this.value,
    this.animationValue,
    @required this.textDirection,
  }) : assert(textDirection != null);

  final Color backgroundColor;
  final Color valueColor;
  final double value;
  final double animationValue;
  final TextDirection textDirection;

  // The indeterminate progress animation displays two lines whose leading (head)
  // and trailing (tail) endpoints are defined by the following four curves.
  static const Curve line1Head = Interval(
    0.0,
    750.0 / _kIndeterminateLinearDuration,
    curve: Cubic(0.2, 0.0, 0.8, 1.0),
  );
  static const Curve line1Tail = Interval(
    333.0 / _kIndeterminateLinearDuration,
    (333.0 + 750.0) / _kIndeterminateLinearDuration,
    curve: Cubic(0.4, 0.0, 1.0, 1.0),
  );
  static const Curve line2Head = Interval(
    1000.0 / _kIndeterminateLinearDuration,
    (1000.0 + 567.0) / _kIndeterminateLinearDuration,
    curve: Cubic(0.0, 0.0, 0.65, 1.0),
  );
  static const Curve line2Tail = Interval(
    1267.0 / _kIndeterminateLinearDuration,
    (1267.0 + 533.0) / _kIndeterminateLinearDuration,
    curve: Cubic(0.10, 0.0, 0.45, 1.0),
  );

  @override
  void paint(Canvas canvas, Size size) {
    final paint = Paint()
      ..color = backgroundColor
      ..style = PaintingStyle.fill;
    canvas.drawRect(Offset.zero & size, paint);

    paint.color = valueColor;

    void drawBar(double x, double width) {
      if (width <= 0.0) {
        return;
      }

      double left;
      switch (textDirection) {
        case TextDirection.rtl:
          left = size.width - width - x;
          break;
        case TextDirection.ltr:
          left = x;
          break;
      }
      canvas.drawRect(Offset(left, 0.0) & Size(width, size.height), paint);
    }

    if (value != null) {
      drawBar(0.0, value.clamp(0.0, 1.0) * size.width);
    } else {
      final x1 = size.width * line1Tail.transform(animationValue);
      final width1 = size.width * line1Head.transform(animationValue) - x1;

      final x2 = size.width * line2Tail.transform(animationValue);
      final width2 = size.width * line2Head.transform(animationValue) - x2;

      drawBar(x1, width1);
      drawBar(x2, width2);
    }
  }

  @override
  bool shouldRepaint(_StepperProgressIndicatorPainter oldPainter) {
    return oldPainter.backgroundColor != backgroundColor ||
        oldPainter.valueColor != valueColor ||
        oldPainter.value != value ||
        oldPainter.animationValue != animationValue ||
        oldPainter.textDirection != textDirection;
  }
}
