import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

import 'form_input_base.dart';

class NumFormInput extends FormInputBase<num> {
  NumFormInput(
      {Key key,
      this.step,
      this.fractionDigits = 0,
      this.minValue = 0,
      this.maxValue,
      this.valueController,
      String labelText,
      String hintText,
      num initialValue,
      this.focusNode,
      bool readOnly = false,
      bool enabled = true,
      bool isRequired = false,
      TextInputAction textInputAction,
      TextAlign textAlign = TextAlign.start,
      VoidCallback onEditingComplete,
      ValueChanged<num> onFieldSubmitted,
      FormFieldSetter<num> onSaved,
      ValueValidator<num> validator,
      ValueChanged<num> onChanged,
      AutovalidateMode autovalidateMode,
      TextAlign labelAlign = TextAlign.left,
      FieldOrientation orientation = FieldOrientation.vertical,
      EdgeInsetsGeometry contentPadding = const EdgeInsets.all(8),
      EdgeInsetsGeometry margin = const EdgeInsets.symmetric(vertical: 8),
      ControlPosition controlPosition = ControlPosition.trailing})
      : assert(initialValue == null || valueController == null),
        super(
          initialValue: initialValue,
          onSaved: onSaved,
          validator: (value, errorCode, errorMsg) {
            String eCode = errorCode;
            String eMsg = errorMsg;
            if (errorCode == null || errorCode.isEmpty) {
              if (minValue != null && minValue > value) {
                eCode = 'lessThanMinValue';
                eMsg =
                    '${labelText ?? "This field"} cannot be less than $minValue.';
              } else if (maxValue != null && maxValue < value) {
                eCode = 'greaterThanMaxValue';
                eMsg =
                    '${labelText ?? "This field"} cannot be greater than $maxValue.';
              }
            }
            if (validator != null) return validator(value, eCode, eMsg);
            return eMsg;
          },
          enabled: enabled,
          autovalidateMode: autovalidateMode,
          orientation: orientation,
          controlPosition: controlPosition,
          labelText: labelText,
          labelAlign: labelAlign,
          valueController: valueController,
          isRequired: isRequired,
          contentPadding: contentPadding,
          builder: (field) {
            NumFormInput widget = field.widget;
            final _CounterFormFieldState state = field;
            final theme = Theme.of(field.context);
            final child = Row(
              children: <Widget>[
                Expanded(
                  child: TextField(
                    inputFormatters: [
                      FilteringTextInputFormatter.allow(RegExp("[0-9.]")),
                    ],
                    controller: state._textEditingController,
                    focusNode: state.effectiveFocusNode,
                    decoration: null,
                    keyboardType: TextInputType.number,
                    textInputAction: textInputAction,
                    textAlign: textAlign,
                    textCapitalization: TextCapitalization.none,
                    autofocus: false,
                    autocorrect: true,
                    onEditingComplete: onEditingComplete,
                    enabled: enabled,
                  ),
                ),
                if (step != null)
                  widget.buildIcon(theme, Icons.remove, onTap: () {
                    state.changeValue((state.value ?? 0) - step);
                  }),
                if (step != null)
                  widget.buildIcon(theme, Icons.add, onTap: () {
                    state.changeValue((state.value ?? 0) + step);
                  }),
              ],
            );
            return widget.buildField(field, child, hasWhiteBackgroud: true);
          },
        );

  /// Controls the text being edited.
  ///
  /// If null, this widget will create its own [TextEditingController] and
  /// initialize its [TextEditingController.text] with [initialValue].
  final num step;
  final int fractionDigits;
  final num minValue;
  final num maxValue;
  final ValueController<num> valueController;
  final FocusNode focusNode;

  @override
  _CounterFormFieldState createState() => _CounterFormFieldState();
}

class _CounterFormFieldState extends FormInputStateBase<num> {
  TextEditingController _textEditingController;
  FocusNode _focusNode;

  @protected
  FocusNode get effectiveFocusNode => widget.focusNode ?? _focusNode;

  @override
  void initState() {
    super.initState();
    _textEditingController = TextEditingController(text: getText());
    if (widget.focusNode == null) {
      _focusNode = FocusNode();
    }

    effectiveFocusNode.addListener(_focusChanged);
  }

  void _focusChanged() {
    if (!effectiveFocusNode.hasFocus) {
      final text = _textEditingController.text;
      num v;
      if (text != null && text.isNotEmpty) {
        v = num.tryParse(text);
        v = v ?? value;
      }
      changeValue(v);
    }
  }

  @override
  void didUpdateWidget(NumFormInput oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.focusNode != oldWidget.focusNode) {
      oldWidget.focusNode?.removeListener(_focusChanged);
      widget.focusNode?.addListener(_focusChanged);

      if (oldWidget.focusNode != null && widget.focusNode == null) {
        _focusNode = FocusNode();
      }
      if (widget.focusNode != null) {
        if (oldWidget.focusNode == null) _focusNode = null;
      }
    }
  }

  String getText() {
    return effectiveValueController.value
            ?.toStringAsFixed(widget.fractionDigits) ??
        '';
  }

  @override
  NumFormInput get widget => super.widget;

  @override
  void dispose() {
    widget.focusNode?.removeListener(_focusChanged);
    _focusNode?.removeListener(_focusChanged);
    _focusNode?.dispose();
    _textEditingController.dispose();
    super.dispose();
  }

  @override
  void changeValue(num v) {
    super.changeValue(v);
    final text = getText();
    if(_textEditingController.text!=text){
      _textEditingController.text=text;
    }
  }
}
//
//class NumericTextFormatter extends TextInputFormatter {
//  NumericTextFormatter({this.fractionDigits = 0});
//
//  int fractionDigits;
//
//  TextEditingValue formatEditUpdate(
//      TextEditingValue oldValue, TextEditingValue newValue) {
//    if (newValue.text.isEmpty) {
//      return newValue.copyWith(text: '');
//    } else if (newValue.text.compareTo(oldValue.text) != 0) {
//      var selectionIndexFromTheRight =
//          newValue.text.length - newValue.selection.end;
//      var s = '';
//      for (var i = 0; i < fractionDigits; i++) {
//        s += '#';
//      }
//      final f = NumberFormat("#,###${s.isNotEmpty ? '.$s' : ''}");
//      var num = int.parse(newValue.text.replaceAll(f.symbols.GROUP_SEP, ''));
//      final newString = f.format(num);
//      return TextEditingValue(
//        text: newString,
//        selection: TextSelection.collapsed(
//            offset: newString.length - selectionIndexFromTheRight),
//      );
//    } else {
//      return newValue;
//    }
//  }
//}
