import 'package:flutter/material.dart';
import 'package:validators/validators.dart';

import '../../mobile_ui.dart';
import 'form_input_base.dart';

enum TextType { text, password, email, url, phone, ip, money }

class TextFormInput extends FormInputBase<String> {
  final ValueController<String> valueController;
  final FocusNode focusNode;
  final Icon icon;
  final TextEditingController textController;

  TextFormInput(
      {Key key,
      String initialValue,
      this.valueController,
      this.textController,
      String labelText,
      String hintText,
      int minLength,
      int maxLength,
      bool isRequired = false,
      bool isShowCounter,
      TextType textType = TextType.text,
      bool readOnly = false,
      bool enabled = true,
      FormFieldSetter<String> onSaved,
      ValueChanged<String> onChanged,
      ValueChanged<String> onSubmitted,
      VoidCallback onEditingComplete,
      this.focusNode,
      TextInputAction textInputAction,
      this.icon,
      bool autofocus = false,
      ValueValidator<String> validator,
      int maxLines = 1,
      TextAlign textAlign = TextAlign.start,
      AutovalidateMode autovalidateMode,
      TextAlign labelAlign = TextAlign.left,
      FieldOrientation orientation = FieldOrientation.vertical,
      EdgeInsetsGeometry contentPadding = const EdgeInsets.all(8),
      EdgeInsetsGeometry margin = const EdgeInsets.symmetric(vertical: 8),
      ControlPosition controlPosition = ControlPosition.trailing})
      : assert((initialValue == null &&
                (valueController == null || textController == null)) ||
            initialValue != null &&
                valueController == null &&
                textController == null),
        assert(maxLines != null && maxLines >= 1),
        super(
          key: key,
          initialValue: initialValue,
          onSaved: onSaved,
          validator: (value, errorCode, errorMsg) {
            String eCode = errorCode;
            String eMsg = errorMsg;
            if (errorCode == null || errorCode.isEmpty) {
              if (minLength != null && value.length < minLength) {
                eCode = 'lengthNotEnough';
                eMsg = '${labelText ?? "This field"} is required.';
              } else if (maxLength != null && value.length > maxLength) {
                eCode = 'lengthExceeded';
                eMsg =
                    '${labelText ?? "This field"} cannot have less than $minLength characters.';
              } else if (textType == TextType.email && !isEmail(value)) {
                eCode = 'invalidEmail';
                eMsg =
                    '${labelText ?? "This field"} requires a valid email address.';
              } else if (textType == TextType.url && !isURL(value)) {
                eCode = 'invalidUrl';
                eMsg =
                    '${labelText ?? "This field"} requires a valid url address.';
              } else if (textType == TextType.ip && !isIP(value)) {
                eCode = 'invalidIp';
                eMsg =
                    '${labelText ?? "This field"} requires a valid ip address.';
              } else if (textType == TextType.phone && !isNumeric(value)) {
                eCode = 'invalidPhone';
                eMsg =
                    '${labelText ?? "This field"} requires a valid phone number.';
              }
            }
            if (validator != null) return validator(value, eCode, eMsg);
            return eMsg;
          },
          enabled: enabled,
          autovalidateMode: autovalidateMode,
          orientation: orientation,
          controlPosition: controlPosition,
          labelText: labelText,
          labelAlign: labelAlign,
          valueController: valueController,
          isRequired: isRequired,
          contentPadding: contentPadding,
          builder: (field) {
            TextFormInput widget = field.widget;
            final _TextFormFieldState state = field;
            final theme = Theme.of(field.context);
            var obscureText = false;
            TextInputType keyboardType;
            switch (textType) {
              case TextType.password:
                {
                  keyboardType = TextInputType.visiblePassword;
                  obscureText = true;
                }
                break;
              case TextType.email:
                {
                  keyboardType = TextInputType.emailAddress;
                }
                break;
              case TextType.url:
                {
                  keyboardType = TextInputType.url;
                }
                break;
              case TextType.phone:
                {
                  keyboardType = TextInputType.phone;
                }
                break;
              case TextType.ip:
                {
                  keyboardType = TextInputType.number;
                }
                break;
              case TextType.money:
                {
                  keyboardType = TextInputType.numberWithOptions(
                      decimal: true, signed: true);
                }
                break;
              default:
                {
                  keyboardType = TextInputType.text;
                }
                break;
            }
            Widget builtCounter;
            bool _isShowCounter = isShowCounter ?? maxLines > 1;
            if (maxLength != null && _isShowCounter) {
              var style = enabled
                  ? theme.textTheme.caption
                  : theme.textTheme.caption
                      .copyWith(color: theme.disabledColor);
              builtCounter = Align(
                alignment: Alignment.bottomRight,
                child: Text(
                  "${state._effectiveController.text.length} / $maxLength",
                  style: style,
                ),
              );
            }
            final child = StatefulBuilder(builder: (context, setState) {
              if (maxLines > 1) {
                keyboardType = TextInputType.multiline;
              }
              final textField = Row(children: <Widget>[
                Expanded(
                    child: TextField(
                  controller: state._effectiveController,
                  autofocus: autofocus,
                  focusNode: focusNode,
                  maxLines: maxLines,
                  obscureText: obscureText,
                  readOnly: readOnly,
                  enabled: enabled,
                  decoration: null,
                  textInputAction: textInputAction ??
                      (maxLines > 1
                          ? TextInputAction.newline
                          : TextInputAction.next),
                  keyboardType: keyboardType,
                  textAlign: textAlign,
                  textCapitalization: TextCapitalization.none,
                  autocorrect: true,
                  onEditingComplete: onEditingComplete,
                  onSubmitted: onSubmitted,
                  onChanged: (String value) {
                    if (onChanged != null) {
                      onChanged(value);
                    }
                    field.didChange(value);
                  },
                  maxLength: maxLength,
                )),
                if (textType == TextType.password)
                  widget.buildIcon(
                      theme,
                      obscureText == true
                          ? Icons.visibility
                          : Icons.visibility_off, onTap: () {
                    if (!enabled) return;
                    setState(
                        () => obscureText = obscureText == true ? false : true);
                  }),
              ]);
              if (builtCounter != null)
                return Column(
                  children: [textField, if (builtCounter != null) builtCounter],
                );
              else
                return textField;
            });
            return widget.buildField(field, child, hasWhiteBackgroud: true);
          },
        );

  @override
  _TextFormFieldState createState() => _TextFormFieldState();
}

class _TextFormFieldState extends FormInputStateBase<String> {
  TextEditingController _controller;

  TextEditingController get _effectiveController =>
      widget.textController ?? _controller;

  @override
  void initState() {
    super.initState();
    if (widget.textController == null) {
      _controller = TextEditingController(text: widget.initialValue);
    } else {
      widget.textController.addListener(_handleControllerChanged);
    }
  }

  @override
  void didUpdateWidget(TextFormInput oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.textController != oldWidget.textController) {
      oldWidget.textController?.removeListener(_handleControllerChanged);
      widget.textController?.addListener(_handleControllerChanged);

      if (oldWidget.textController != null && widget.textController == null)
        _controller =
            TextEditingController.fromValue(oldWidget.textController.value);
      if (widget.textController != null) {
        setValue(widget.textController.text);
        if (oldWidget.textController == null) _controller = null;
      }
    }
  }

  @override
  TextFormInput get widget => super.widget;

  @override
  void dispose() {
    widget.textController?.removeListener(_handleControllerChanged);
    super.dispose();
  }

  @override
  void changeValue(String v) {
    super.changeValue(v);
    if (_effectiveController.text != v) {
      _effectiveController.text = v;
    }
  }

  void _handleControllerChanged() {
    if (_effectiveController.text != value)
      changeValue(_effectiveController.text);
  }
}
